(*
    This file is part of the FunctionalIO project -
    which provides functional input streams.
    
    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

(* Notice: This is not thread safe! *)
functor FFunctionalIO(type vec
                      type element
                      type instream
                      val impOpenIn : string -> instream
                      val impCloseIn : instream -> unit
                      val impInput : instream -> vec
                      val vecConcat : vec list -> vec
                      val vecExtract : (vec * int * int option) -> vec
                      val vecSub : (vec * int) -> element
                      val vecLength : vec -> int) =
  struct
    open IO

    type vector = vec
    type elem = element

    datatype impInChunk =
      ImpInChunkEnd
    | ImpInChunkVector of vector * (impInChunk ref)
    | ImpInChunkStream of instream

    (* The integer is a functional position in the first
       vector chunk - if there is a vector chunk at all. *)
    type instream = int * (impInChunk ref)
    type outstream = unit

    (* Internal function implementing inputN *)
    fun fInputN _ acc inStrRef ImpInChunkEnd n =
          (* (print "inputN, ChunkEnd\n"; *)
          (vecConcat (rev acc), (0, inStrRef))
      | fInputN pos acc inStrRef (ImpInChunkVector (v, next)) n =
          let
            (* val _ = print "inputN, ChunkVector\n" *)
            val len = vecLength v - pos
          in
            if len = n then
              if pos = 0 then
                (vecConcat (rev (v::acc)), (0, next))
              else
                let
                  val data = vecExtract (v, pos, SOME(n)) 
                in
                  (vecConcat (rev (data::acc)), (0, next))
                end
            else if len > n then
              let
                val data = vecExtract (v, pos, SOME(n))
              in
                (vecConcat (rev (data::acc)),
                 (pos + n, inStrRef))
              end
            else (* i.e.: len < n *)
              let
                val data = vecExtract (v, pos, SOME(len))
              in
                fInputN 0 (data::acc) next (!next) (n - len)
              end
          end
      | fInputN _ acc inStrRef (ImpInChunkStream inStr) n =
          let
            (* val _ = print "inputN, ChunkStream\n" *)
            val newVec = impInput inStr
            val _ = inStrRef :=
                      (case vecLength newVec of
                         0 => (impCloseIn inStr;
                               ImpInChunkEnd)
                       | _ => ImpInChunkVector
                                (newVec, ref (ImpInChunkStream inStr)))
          in
            fInputN 0 acc inStrRef (!inStrRef) n
          end
          

    (* FIXME: Return 0-length vector for NONE? *)
    fun inputN ((pos, inStr), n) =
          if n < 0 then (* FIXME: No check for maxLen! *)
            raise Size
          else
            fInputN pos nil inStr (!inStr) n

    (* FIXME: Implement more efficiently... *)
    fun input1 inStr =
          let
            val (v, s) = inputN (inStr, 1)
          in
            if vecLength v >= 1 then
              SOME(vecSub (v, 0), s)
            else
              NONE
          end

    (* Internal function implementing input *)
    fun fInput pos inStrRef (ImpInChunkVector (v, next)) =
          (vecExtract (v, pos, NONE), (0, next))
      | fInput pos inStrRef _ =
          (* FIXME: Implement more efficiently? *)
          fInputN 0 nil inStrRef (!inStrRef) 64

    fun input (pos, inStr) =
          fInput pos inStr (!inStr)

    (* Internal function implementing closeIn
       Maybe we should make a more general parametrized
       traverse function... *)
    fun fCloseIn inStrRef (ImpInChunkStream inStr) =
          (impCloseIn inStr;
           inStrRef := ImpInChunkEnd)
      | fCloseIn _ ImpInChunkEnd =
          ()
      | fCloseIn _ (ImpInChunkVector (_, next)) =
          fCloseIn next (!next)

    (* Closes file for further input.
       New end of file becomes the furthest
       position in the file that has been read
       internally. *)
    fun closeIn (_, inStr) =
          fCloseIn inStr (!inStr)

    fun openIn fileName =
          (0,
           ref (ImpInChunkStream
                  (impOpenIn fileName)))
  end

structure FuncBinIO =
  FFunctionalIO(type vec = Word8Vector.vector
                type element = Word8.word
                type instream = BinIO.instream
                val impOpenIn = BinIO.openIn
                val impCloseIn = BinIO.closeIn
                val impInput = BinIO.input
                val vecConcat = Word8Vector.concat
                val vecExtract = Word8Vector.extract
                val vecSub = Word8Vector.sub
                val vecLength = Word8Vector.length)
    :> FUNC_BIN_IO
         where type vector = Word8Vector.vector
           and type elem = Word8.word

structure FuncTextIO =
  FFunctionalIO(type vec = CharVector.vector
                type element = Char.char
                type instream = TextIO.instream
                val impOpenIn = TextIO.openIn
                val impCloseIn = TextIO.closeIn
                val impInput = TextIO.input
                val vecConcat = CharVector.concat
                val vecExtract = CharVector.extract
                val vecSub = CharVector.sub
                val vecLength = CharVector.length)
    :> FUNC_TEXT_IO
         where type vector = CharVector.vector
           and type elem = Char.char